@@ -1,18 +1,15 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * This file is part of libgpiod.
  *
  * Copyright (C) 2017-2018 Bartosz Golaszewski <bartekgola@gmail.com>
- *
- * This program is free software; you can redistribute it and/or modify it
- * under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 2.1 of the License, or (at
- * your option) any later version.
  */
 
 /* Low-level, core library code. */
 
 #include <gpiod.h>
 
+#include <stdio.h>
 #include <string.h>
 #include <stdint.h>
 #include <unistd.h>
@@ -22,6 +19,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <sys/sysmacros.h>
 #include <linux/gpio.h>
 
 enum {
@@ -63,6 +61,86 @@ struct gpiod_chip {
 	char label[32];
 };
 
+static bool is_gpiochip_cdev(const char *path)
+{
+	char *name, *pathcpy, *sysfsp, sysfsdev[16], devstr[16];
+	struct stat statbuf;
+	bool ret = false;
+	int rv, fd;
+	ssize_t rd;
+
+	rv = lstat(path, &statbuf);
+	if (rv)
+		goto out;
+
+	/* Is it a character device? */
+	if (!S_ISCHR(statbuf.st_mode)) {
+		/*
+		 * Passing a file descriptor not associated with a character
+		 * device to ioctl() makes it set errno to ENOTTY. Let's do
+		 * the same in order to stay compatible with the versions of
+		 * libgpiod from before the introduction of this routine.
+		 */
+		errno = ENOTTY;
+		goto out;
+	}
+
+	/* Get the basename. */
+	pathcpy = strdup(path);
+	if (!pathcpy)
+		goto out;
+
+	name = basename(pathcpy);
+
+	/* Do we have a corresponding sysfs attribute? */
+	rv = asprintf(&sysfsp, "/sys/bus/gpio/devices/%s/dev", name);
+	if (rv < 0)
+		goto out_free_pathcpy;
+
+	if (access(sysfsp, R_OK) != 0) {
+		/*
+		 * This is a character device but not the one we're after.
+		 * Before the introduction of this function, we'd fail with
+		 * ENOTTY on the first GPIO ioctl() call for this file
+		 * descriptor. Let's stay compatible here and keep returning
+		 * the same error code.
+		 */
+		errno = ENOTTY;
+		goto out_free_sysfsp;
+	}
+
+	/*
+	 * Make sure the major and minor numbers of the character device
+	 * correspond with the ones in the dev attribute in sysfs.
+	 */
+	snprintf(devstr, sizeof(devstr), "%u:%u",
+		 major(statbuf.st_rdev), minor(statbuf.st_rdev));
+
+	fd = open(sysfsp, O_RDONLY);
+	if (fd < 0)
+		goto out_free_sysfsp;
+
+	memset(sysfsdev, 0, sizeof(sysfsdev));
+	rd = read(fd, sysfsdev, strlen(devstr));
+	close(fd);
+	if (rd < 0)
+		goto out_free_sysfsp;
+
+	if (strcmp(sysfsdev, devstr) != 0) {
+		errno = ENODEV;
+		goto out_free_sysfsp;
+	}
+
+	ret = true;
+
+out_free_sysfsp:
+	free(sysfsp);
+out_free_pathcpy:
+	free(pathcpy);
+out:
+	return ret;
+}
+
 struct gpiod_chip *gpiod_chip_open(const char *path)
 {
 	struct gpiochip_info info;
@@ -73,6 +151,15 @@ struct gpiod_chip *gpiod_chip_open(const char *path)
 	if (fd < 0)
 		return NULL;
 
+	/*
+	 * We were able to open the file but is it really a gpiochip character
+	 * device?
+	 */
+	if (!is_gpiochip_cdev(path)) {
+		close(fd);
+		return NULL;
+	}
+
 	chip = malloc(sizeof(*chip));
 	if (!chip)
 		goto err_close_fd;
@@ -415,10 +502,11 @@ static int line_request_values(struct gpiod_line_bulk *bulk,
 
 	req.lines = gpiod_line_bulk_num_lines(bulk);
 
-	for (i = 0; i < gpiod_line_bulk_num_lines(bulk); i++) {
-		line = gpiod_line_bulk_get_line(bulk, i);
+	gpiod_line_bulk_foreach_line_off(bulk, line, i) {
 		req.lineoffsets[i] = gpiod_line_offset(line);
-		if (config->request_type == GPIOD_LINE_REQUEST_DIRECTION_OUTPUT)
+		if (config->request_type ==
+				GPIOD_LINE_REQUEST_DIRECTION_OUTPUT &&
+		    default_vals)
 			req.default_values[i] = !!default_vals[i];
 	}
 
@@ -701,16 +789,23 @@ int gpiod_line_event_wait_bulk(struct gpiod_line_bulk *bulk,
 	else if (rv == 0)
 		return 0;
 
-	if (event_bulk) {
+	if (event_bulk)
 		gpiod_line_bulk_init(event_bulk);
 
-		for (off = 0; off < num_lines; off++) {
-			if (fds[off].revents) {
+	for (off = 0; off < num_lines; off++) {
+		if (fds[off].revents) {
+			if (fds[off].revents & POLLNVAL) {
+				errno = EINVAL;
+				return -1;
+			}
+
+			if (event_bulk) {
 				line = gpiod_line_bulk_get_line(bulk, off);
 				gpiod_line_bulk_add(event_bulk, line);
-				if (!--rv)
-					break;
 			}
+
+			if (!--rv)
+				break;
 		}
 	}
 
